@@ -1,7 +1,9 @@
 package aghtls
 
 import (
+	"context"
 	"crypto/x509"
+	"log/slog"
 )
 
 // SystemRootCAs tries to load root certificates from the operating system.  It
@@ -9,6 +11,6 @@ import (
 // default algorithm to find system root CA list.
 //
 // See https://github.com/AdguardTeam/AdGuardHome/issues/1311.
-func SystemRootCAs() (roots *x509.CertPool) {
-	return rootCAs()
+func SystemRootCAs(ctx context.Context, l *slog.Logger) (roots *x509.CertPool) {
+	return rootCAs(ctx, l)
 }
