@@ -1,39 +1,41 @@
 package dnsforward
 
 import (
+	"context"
 	"net"
-	"strings"
 	"time"
 
+	"github.com/AdguardTeam/AdGuardHome/internal/aghnet"
 	"github.com/AdguardTeam/AdGuardHome/internal/filtering"
 	"github.com/AdguardTeam/AdGuardHome/internal/querylog"
 	"github.com/AdguardTeam/AdGuardHome/internal/stats"
 	"github.com/AdguardTeam/dnsproxy/proxy"
-	"github.com/AdguardTeam/golibs/log"
-	"github.com/AdguardTeam/golibs/netutil"
 	"github.com/miekg/dns"
-	"golang.org/x/exp/slices"
 )
 
 // Write Stats data and logs
-func (s *Server) processQueryLogsAndStats(dctx *dnsContext) (rc resultCode) {
-	log.Debug("dnsforward: started processing querylog and stats")
-	defer log.Debug("dnsforward: finished processing querylog and stats")
+func (s *Server) processQueryLogsAndStats(ctx context.Context, dctx *dnsContext) (rc resultCode) {
+	s.logger.DebugContext(ctx, "started processing querylog and stats")
+	defer s.logger.DebugContext(ctx, "finished processing querylog and stats")
 
-	elapsed := time.Since(dctx.startTime)
 	pctx := dctx.proxyCtx
-
 	q := pctx.Req.Question[0]
-	host := strings.ToLower(strings.TrimSuffix(q.Name, "."))
+	host := aghnet.NormalizeDomain(q.Name)
+	processingTime := time.Since(dctx.startTime)
 
-	ip, _ := netutil.IPAndPortFromAddr(pctx.Addr)
-	ip = slices.Clone(ip)
+	ip := pctx.Addr.Addr().AsSlice()
 	s.anonymizer.Load()(ip)
+	ipStr := net.IP(ip).String()
 
-	log.Debug("dnsforward: client ip for stats and querylog: %s", ip)
+	s.logger.DebugContext(ctx, "client ip for stats and querylog", "ip", ipStr)
+
+	ids := []string{ipStr}
+	if dctx.clientID != "" {
+		// Use the ClientID first because it has a higher priority.  Filters
+		// have the same priority, see applyAdditionalFiltering.
+		ids = []string{dctx.clientID, ipStr}
+	}
 
-	ipStr := ip.String()
-	ids := []string{ipStr, dctx.clientID}
 	qt, cl := q.Qtype, q.Qclass
 
 	// Synchronize access to s.queryLog and s.stats so they won't be suddenly
@@ -43,26 +45,28 @@ func (s *Server) processQueryLogsAndStats(dctx *dnsContext) (rc resultCode) {
 	defer s.serverLock.RUnlock()
 
 	if s.shouldLog(host, qt, cl, ids) {
-		s.logQuery(dctx, pctx, elapsed, ip)
+		s.logQuery(dctx, ip, processingTime)
 	} else {
-		log.Debug(
-			"dnsforward: request %s %s %q from %s ignored; not adding to querylog",
-			dns.Class(cl),
-			dns.Type(qt),
-			host,
-			ip,
+		s.logger.DebugContext(
+			ctx,
+			"not adding to querylog",
+			"dns_class", dns.Class(cl),
+			"dns_type", dns.Type(qt),
+			"host", host,
+			"ip", ipStr,
 		)
 	}
 
 	if s.shouldCountStat(host, qt, cl, ids) {
-		s.updateStats(dctx, elapsed, *dctx.result, ipStr)
+		s.updateStats(dctx, ipStr, processingTime)
 	} else {
-		log.Debug(
-			"dnsforward: request %s %s %q from %s ignored; not counting in stats",
-			dns.Class(cl),
-			dns.Type(qt),
-			host,
-			ip,
+		s.logger.DebugContext(
+			ctx,
+			"not counting in stats",
+			"dns_class", dns.Class(cl),
+			"dns_type", dns.Type(qt),
+			"host", host,
+			"ip", ipStr,
 		)
 	}
 
@@ -90,12 +94,9 @@ func (s *Server) shouldCountStat(host string, qt, cl uint16, ids []string) (ok b
 }
 
 // logQuery pushes the request details into the query log.
-func (s *Server) logQuery(
-	dctx *dnsContext,
-	pctx *proxy.DNSContext,
-	elapsed time.Duration,
-	ip net.IP,
-) {
+func (s *Server) logQuery(dctx *dnsContext, ip net.IP, processingTime time.Duration) {
+	pctx := dctx.proxyCtx
+
 	p := &querylog.AddParams{
 		Question:          pctx.Req,
 		ReqECS:            pctx.ReqECS,
@@ -104,7 +105,7 @@ func (s *Server) logQuery(
 		Result:            dctx.result,
 		ClientID:          dctx.clientID,
 		ClientIP:          ip,
-		Elapsed:           elapsed,
+		Elapsed:           processingTime,
 		AuthenticatedData: dctx.responseAD,
 	}
 
@@ -123,46 +124,52 @@ func (s *Server) logQuery(
 
 	if pctx.Upstream != nil {
 		p.Upstream = pctx.Upstream.Address()
-	} else if cachedUps := pctx.CachedUpstreamAddr; cachedUps != "" {
-		p.Upstream = pctx.CachedUpstreamAddr
-		p.Cached = true
+	}
+
+	if qs := pctx.QueryStatistics(); qs != nil {
+		ms := qs.Main()
+		if len(ms) == 1 && ms[0].IsCached {
+			p.Upstream = ms[0].Address
+			p.Cached = true
+		}
 	}
 
 	s.queryLog.Add(p)
 }
 
-// updatesStats writes the request into statistics.
-func (s *Server) updateStats(
-	ctx *dnsContext,
-	elapsed time.Duration,
-	res filtering.Result,
-	clientIP string,
-) {
-	pctx := ctx.proxyCtx
-	e := stats.Entry{}
-	e.Domain = strings.ToLower(pctx.Req.Question[0].Name)
-	if e.Domain != "." {
-		// Remove last ".", but save the domain as is for "." queries.
-		e.Domain = e.Domain[:len(e.Domain)-1]
+// updateStats writes the request data into statistics.
+func (s *Server) updateStats(dctx *dnsContext, clientIP string, processingTime time.Duration) {
+	pctx := dctx.proxyCtx
+
+	var upstreamStats []*proxy.UpstreamStatistics
+	qs := pctx.QueryStatistics()
+	if qs != nil {
+		upstreamStats = append(upstreamStats, qs.Main()...)
+		upstreamStats = append(upstreamStats, qs.Fallback()...)
 	}
 
-	if clientID := ctx.clientID; clientID != "" {
+	e := &stats.Entry{
+		UpstreamStats:  upstreamStats,
+		Domain:         aghnet.NormalizeDomain(pctx.Req.Question[0].Name),
+		Result:         stats.RNotFiltered,
+		ProcessingTime: processingTime,
+	}
+
+	if clientID := dctx.clientID; clientID != "" {
 		e.Client = clientID
 	} else {
 		e.Client = clientIP
 	}
 
-	e.Time = uint32(elapsed / 1000)
-	e.Result = stats.RNotFiltered
-
-	switch res.Reason {
+	switch dctx.result.Reason {
 	case filtering.FilteredSafeBrowsing:
 		e.Result = stats.RSafeBrowsing
 	case filtering.FilteredParental:
 		e.Result = stats.RParental
 	case filtering.FilteredSafeSearch:
 		e.Result = stats.RSafeSearch
-	case filtering.FilteredBlockList,
+	case
+		filtering.FilteredBlockList,
 		filtering.FilteredInvalid,
 		filtering.FilteredBlockedService:
 		e.Result = stats.RFiltered
